/*
	*
	* Copyright 2020 Britanicus <marcusbritanicus@gmail.com>
	*
	* This file is a part of DesQ project (https://gitlab.com/desq/)
	*

	*
	* This program is free software; you can redistribute it and/or modify
	* it under the terms of the GNU General Public License as published by
	* the Free Software Foundation; either version 3 of the License, or
	* (at your option) any later version.
	*

	*
	* This program is distributed in the hope that it will be useful,
	* but WITHOUT ANY WARRANTY; without even the implied warranty of
	* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	* GNU General Public License for more details.
	*

	*
	* You should have received a copy of the GNU General Public License
	* along with this program; if not, write to the Free Software
	* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
	* MA 02110-1301, USA.
	*
*/

#include <QDebug>
#include <ctime>

#include "DesQShellManager.hpp"
#include "DesQAutoStart.hpp"
#include "DesQShellAdaptor.hpp"

#include <desq/DesQUtils.hpp>
#include <desq/DesQSettings.hpp>
#include <desq/desq-config.h>

#include <desqwl/DesQWlGlobal.hpp>
#include <desqwl/DesQRegistry.hpp>
#include <desqwl/DesQHotSpot.hpp>

int MaxAutoRestartLimit;

DesQShellManager::DesQShellManager() : QObject() {

	shellSett = new DesQSettings( "Shell" );
	MaxAutoRestartLimit = shellSett->value( "Manager/MaxAutoRestartLimit" );

	qDebug( "Connecting to DesQ Session" );
	session = new QDBusInterface( "org.DesQ.Session", "/org/DesQ/Session", "org.DesQ.Session", QDBusConnection::sessionBus() );
	new ShellAdaptor( this );

	qDebug( "Registering DesQ Shell Service" );
	QDBusConnection shellDBus = QDBusConnection::sessionBus();
    serviceRunning = shellDBus.registerService( "org.DesQ.Shell" );
    objectRegistered = shellDBus.registerObject( "/org/DesQ/Shell", this );

	session->call( "RegisterService", "org.DesQ.Shell" );
};

DesQShellManager::~DesQShellManager() {

	shutdown();
};

bool DesQShellManager::isServiceRunning() {

	return serviceRunning;
};

bool DesQShellManager::isObjectRegistered() {

	return objectRegistered;
};

void DesQShellManager::startManagement() {

	qDebug( "Drawing the Background" );
	bg = new TrackedProcess( DesQUtils::getUtilityPath( "bg" ) );
	bg->start();

	if ( shellSett->value( "ShowPanel" ) ) {
		qDebug( "Opening the Panel" );
		panel = new TrackedProcess( DesQUtils::getUtilityPath( "panel" ) );
		panel->start();
	}

	if ( shellSett->value( "ShowDock" ) ) {
		qDebug( "Loading the Dock" );
		dock = new TrackedProcess( DesQUtils::getUtilityPath( "dock" ) );
		dock->start();
	}

	qDebug( "Creating DesQ Menu" );
	menu = new TrackedProcess( DesQUtils::getUtilityPath( "menu" ) );
	menu->start();

	qDebug( "Adding DesQ Widgets" );
	widgets = new TrackedProcess( DesQUtils::getUtilityPath( "container" ) );
	widgets->start( shellSett->value( "Widgets/EnabledWidgets" ) );

	qDebug( "Starting DesQ SuperStart" );
	// ss = new TrackedProcess( DesQUtils::getUtilityPath( "desq-superstart" ) );
	// ss->start();

    /* Start the DesQ SN Watcher: Registers the SN Items */
	qDebug() << "Starting DesQ SN Watcher";
	QProcess::startDetached( DesQUtils::getUtilityPath( "snwatcher-daemon" ), {} );

    /* Start the DesQ SNI: The DesQ Tray Utility */
	qDebug() << "Starting DesQ SN Tray";
	sni = new TrackedProcess( DesQUtils::getUtilityPath( "sni" ) );
	sni->start();

	// Synthetic 5s delay.
	int t = 0;
	while ( t < 10 ) {
		// Sleep of 500 ms at one time
		// 500 (ms) * 1000 (us per ms) * 1000 (ns per us)
		struct timespec ts = { 0, 500 * 1000 * 1000 };
		nanosleep( &ts, NULL );
		qApp->processEvents();
		t++;
	}

	qDebug() << "Starting auto-start apps";
	DesQAutoStart *autoStart = new DesQAutoStart();
	autoStart->start();

	qDebug() << "Activating hot-corners";
	setupHotCorners();
};

void DesQShellManager::shutdown() {

	delete shellSett;

	if ( bg ) {
		bg->terminate();
		bg->deleteLater();
	}

	if ( panel ) {
		panel->terminate();
		panel->deleteLater();
	}

	if ( dock ) {
		dock->terminate();
		dock->deleteLater();
	}

	if ( menu ) {
		menu->terminate();
		menu->deleteLater();
	}

	if ( sni ) {
		sni->terminate();
		sni->deleteLater();
	}

	if ( ss ) {
		ss->terminate();
		ss->deleteLater();
	}

	if ( widgets ) {
		widgets->terminate();
		widgets->deleteLater();
	}

	if ( session ) {
		session->deleteLater();
	}

	/* Close this daemon */
	qApp->quit();
};

void DesQShellManager::setupHotCorners() {

	wl_display *display = DesQWayland::getWlDisplay();
	if ( not display ) {
		qDebug() << "Unable to acquire wl_display from the compositor.";
		qDebug() << "Your experience will be severly limited.";
	}

	DesQWaylandRegistry *registry = new DesQWaylandRegistry( display );
	QObject::connect(
		registry, &DesQWaylandRegistry::errorOccured, [=]( DesQWaylandRegistry::ErrorType et ) {
			qDebug() << "Error caused on registry" << et;
			qDebug() << "Valiantly trying to continue...";
		}
	);
	registry->initialize();

	DesQHotSpot *topLeftHS = new DesQHotSpot( qApp->primaryScreen(), registry->wayfireShell(), this );
	topLeftHS->createHotspot( DesQHotSpot::TopLeft, 1, 250 );
	connect( topLeftHS, &DesQHotSpot::enteredHotspot, [=]() { handleHotCorner( DesQHotSpot::TopLeft ); } );
	connect( topLeftHS, &DesQHotSpot::leftHotspot,    [=]() { cancelHotCorner( DesQHotSpot::TopLeft ); } );

	DesQHotSpot *topRightHS = new DesQHotSpot( qApp->primaryScreen(), registry->wayfireShell(), this );
	topRightHS->createHotspot( DesQHotSpot::TopRight, 1, 250 );
	connect( topRightHS, &DesQHotSpot::enteredHotspot, [=]() { handleHotCorner( DesQHotSpot::TopRight ); } );
	connect( topRightHS, &DesQHotSpot::leftHotspot,    [=]() { cancelHotCorner( DesQHotSpot::TopRight ); } );

	DesQHotSpot *bottomLeftHS = new DesQHotSpot( qApp->primaryScreen(), registry->wayfireShell(), this );
	bottomLeftHS->createHotspot( DesQHotSpot::BottomLeft, 1, 250 );
	connect( bottomLeftHS, &DesQHotSpot::enteredHotspot, [=]() { handleHotCorner( DesQHotSpot::BottomLeft ); } );
	connect( bottomLeftHS, &DesQHotSpot::leftHotspot,    [=]() { cancelHotCorner( DesQHotSpot::BottomLeft ); } );

	DesQHotSpot *bottomRightHS = new DesQHotSpot( qApp->primaryScreen(), registry->wayfireShell(), this );
	bottomRightHS->createHotspot( DesQHotSpot::BottomRight, 1, 250 );
	connect( bottomRightHS, &DesQHotSpot::enteredHotspot, [=]() { handleHotCorner( DesQHotSpot::BottomRight ); } );
	connect( bottomRightHS, &DesQHotSpot::leftHotspot,    [=]() { cancelHotCorner( DesQHotSpot::BottomRight ); } );
};

void DesQShellManager::handleHotCorner( int corner ) {

	/* Based on the corner take action */
	switch( corner ) {

		case DesQHotSpot::TopLeft: {
			qDebug() << "Entering Top Left hot Corner";
			break;
		}

		case DesQHotSpot::TopRight: {
			qDebug() << "Entering Top Right Corner";
			/* Use TasksPlugin to show the list widget */
			break;
		}

		case DesQHotSpot::BottomLeft: {
			qDebug() << "Entering Bottom Left Corner";
			QDBusInterface iface( "org.DesQ.Menu", "/org/DesQ/Menu", "org.DesQ.Menu", QDBusConnection::sessionBus() );
			iface.call( "Toggle" );

			break;
		}

		case DesQHotSpot::BottomRight: {
			qDebug() << "Entering Bottom Right Corner";
			QProcess::startDetached( DesQUtils::getUtilityPath( "container" ), { "--raise" } );
			break;
		}

		/* In case hot corner was canceled */
		default: {
			break;
		}
	}
};

void DesQShellManager::cancelHotCorner( int corner ) {

	switch( corner ) {

		case DesQHotSpot::TopLeft: {
			qDebug() << "Leaving Top Left Corner";
			break;
		}

		/* We are exiting from the TopRight corner, so show restore the window state */
		case DesQHotSpot::TopRight: {
			qDebug() << "Leaving Top Right Corner";
			break;
		}

		case DesQHotSpot::BottomLeft: {
			qDebug() << "Leaving Bottom Left Corner";
			break;
		}

		case DesQHotSpot::BottomRight: {
			qDebug() << "Leaving Bottom Right Corner";
			QProcess::startDetached( DesQUtils::getUtilityPath( "container" ), { "--lower" } );
			break;
		}

		/* In case hot corner was canceled */
		default: {
			break;
		}
	}
};
